#############################################################################
## Some additional methods of wartslib classes.
##
## --------------------------------------------------------------------------
## Copyright (C) 2007 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: wl-trace.rb,v 1.12 2008/10/02 23:47:08 youngh Exp $
#############################################################################

module Warts

class Trace

  private

  HALT_REASON = []
  HALT_REASON[Warts::Trace::STOP_COMPLETED] = "S"
  HALT_REASON[Warts::Trace::STOP_NONE] = "S"
  HALT_REASON[Warts::Trace::STOP_UNREACH] = "U"
  HALT_REASON[Warts::Trace::STOP_LOOP] = "L"
  HALT_REASON[Warts::Trace::STOP_GAPLIMIT] = "G"

  public

  # from scamper_trace.h:SCAMPER_TRACE_STOP_*
  STOP_REASON_TEXT = Hash.new { |h, k| "<<UNKNOWN STOP CODE #{k}>>" }
  STOP_REASON_TEXT.merge!({
    Warts::Trace::STOP_NONE => "none (null reason)",
    Warts::Trace::STOP_COMPLETED => "completed (got an ICMP port unreach)",
    Warts::Trace::STOP_UNREACH => "unreach (got an other ICMP unreach code)",
    Warts::Trace::STOP_ICMP => "icmp (got an ICMP msg, not unreach)",
    Warts::Trace::STOP_LOOP => "loop (loop detected)",
    Warts::Trace::STOP_GAPLIMIT => "gaplimit (gaplimit exceeded)",
    Warts::Trace::STOP_ERROR => "error (sendto error)",
    Warts::Trace::STOP_HOPLIMIT => "hoplimit (hoplimit reached)"
  })


  # Just a convenience function.  This returns the same strings as warts-dump.
  def self.decompose_tcp_flags(flags)
    retval = []
    retval << "fin" if (flags & 0x01) != 0
    retval << "syn" if (flags & 0x02) != 0
    retval << "rst" if (flags & 0x04) != 0
    retval << "psh" if (flags & 0x08) != 0
    retval << "ack" if (flags & 0x10) != 0
    retval << "urg" if (flags & 0x20) != 0
    retval << "ece" if (flags & 0x40) != 0
    retval << "cwr" if (flags & 0x80) != 0
    retval
  end


  # Just a convenience function.
  # This returns the same strings as warts-dump, except for "gapcont" which
  # arts-dump doesn't return. See scamper_trace.h:SCAMPER_TRACE_FLAG_*
  def self.decompose_flags(flags)
    retval = []
    retval << "all_attempts" if (flags & Warts::Trace::FLAG_ALLATTEMPTS) != 0
    retval << "pmtud" if (flags & Warts::Trace::FLAG_PMTUD) != 0
    retval << "dltxts" if (flags & Warts::Trace::FLAG_DL) != 0
    retval
  end


  # Just a convenience function.
  # This returns the same strings as warts-dump, except for "tcpresp" which
  # arts-dump doesn't return. See scamper_trace.h:SCAMPER_TRACE_HOP_FLAG_*.
  # I commented out printing of REPLY_TTL and TCP because they're mostly of
  # internal interest, and they can be got through accessor methods anyway.
  def self.decompose_hop_flags(flags)
    retval = []
    retval << "sockrxts" if (flags & Warts::Trace::HOP_FLAG_TS_SOCK_RX) != 0
    retval << "dltxts" if (flags & Warts::Trace::HOP_FLAG_TS_DL_TX) != 0
    retval << "dlrxts" if (flags & Warts::Trace::HOP_FLAG_TS_DL_RX) != 0
    retval << "tscrtt" if (flags & Warts::Trace::HOP_FLAG_TS_TSC) != 0
    #retval << "replyttl" if (flags & Warts::Trace::HOP_FLAG_REPLY_TTL) != 0
    #retval << "tcpresp" if (flags & Warts::Trace::HOP_FLAG_TCP) != 0
    retval
  end


  def flags_breakdown
    Warts::Trace.decompose_flags flags
  end


  def hop_flags_breakdown(hop, response=0)
    Warts::Trace.decompose_hop_flags hop_flags(hop, response)
  end


  # The length of the path (in hops), taking any responding destination into
  # account.
  #
  # The hop_count attribute doesn't provide this value because hop_count
  # gives the number of hops probed (that is, attempted) and not the number
  # of hops with responses (or, alternatively, the highest TTL of responding
  # hops).
  def path_length
    dest_response = find_dest_response
    if dest_response
      dest_response[0] + 1
    else
      return 0 if hop_count == 0
      (hop_count-1).downto(0) do |hop|
        return hop + 1 if hop_exists? hop
      end
      return 0
    end
  end

  def dump
    fields = ["T"]

    fields.push src, dest, list_id, cycle_id, start

    dest_response = find_dest_response
    if dest_response
      fields.push "R", hop_rtt_str(*dest_response),
	hop_probe_ttl(*dest_response),
	hop_reply_ttl(*dest_response)
    else
      fields.push "N", 0, 0, 0
    end

    fields.push(HALT_REASON[stop_reason] || "?")
    fields.push stop_data

    fields.push(complete? ? "C" : "I")

    # For compatibility with sc_analysis_dump, we don't print missing hops
    # at the end of the trace.
    unresponsive = 0
    each_hop do |hop, exists|
      if exists
	hop_str = ""
	each_attempt(hop) do |attempt|
	  # sc_analysis_dump compatibility
	  next if dest_response &&
	    hop == dest_response[0] && attempt == dest_response[1]

	  hop_str << ";" unless hop_str.empty?
	  hop_str << hop_addr(hop, attempt)
	  hop_str << "," << hop_rtt_str(hop, attempt)
	  hop_str << "," << hop_probe_id(hop, attempt).to_s
	end

	unless hop_str.empty?
	  unresponsive.times { fields.push "q" }
	  unresponsive = 0

	  fields.push hop_str
	end
      else
	unresponsive += 1
      end
    end

    fields.join("\t")
  end

end

end
